/****
 *
 * Arquivo: FilaC.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/06/2012
 * ltima modificao: 16/06/2012
 *
 * Descrio: Implementao de fila usando array esttico circular
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>      /* Funes de Entrada/Sada */
#include "FilaC.h"      /* Interface deste mdulo   */
#include "ASSEGURA.h"   /* Macro ASSEGURA           */

/***************** Definies de Funes ******************/

/****
 *
 * FilaCheiaC(): Verifica se uma fila circular est repleta
 *
 * Parmetros:
 *      f (entrada) - a fila
 *
 * Retorno: 1, se a fila estiver repleta; 0, em caso contrrio
 *
 * Observao: Esta funo faz parte da implementao de fila e no da abstrao
 *             fila. Por isso, ela deve ser definida como static.
 *
 ****/
static int FilaCheiaC(const tFilaC *f)
{
   return f->frente == (f->fundo + 1) % (MAX_ELEMENTOS + 1);
}

/****
 *
 * CriaFilaC(): Inicia a fila circular recebida como parmetro
 *
 * Parmetros:
 *      f (sada) - a fila a ser iniciada
 *
 * Retorno: Nada
 *
 ****/
void  CriaFilaC(tFilaC *f)
{
   f->frente = f->fundo = 0;
}

/****
 *
 * FilaVaziaC(): Verifica se uma fila circular est vazia
 *
 * Parmetros:
 *      f (entrada) - a fila
 *
 * Retorno: 1, se a fila estiver vazia
 *          0, em caso contrrio
 *
 ****/
int  FilaVaziaC(const tFilaC *f)
{
   return f->frente == f->fundo;
}

/****
 *
 * ElementoFrenteC(): Retorna o elemento que se encontra
 *                    na frente de uma fila circular
 *
 * Parmetros:
 *      f (entrada) - a fila
 *
 * Retorno: O elemento que se encontra na frente da fila
 *
 * Observao: Se a fila estiver vazia, causa o aborto do programa
 *
 ****/
tElemento ElementoFrenteC(const tFilaC *f)
{
   ASSEGURA(!FilaVaziaC(f), "Erro: Fila vazia");

   return f->itens[(f->frente + 1) % (MAX_ELEMENTOS + 1)];
}

/****
 *
 * AcrescentaC(): AcrescentaC um item numa fila circular.
 *
 * Parmetros:
 *      item (entrada) - o item que ser AcrescentaCdo
 *      f (entrada/sada) - a fila na qual ser feita o acrscimo
 *
 * Retorno: Nada.
 *
 * Observao: Se a fila estiver cheia, causa o aborto do programa.
 *
 ****/
void AcrescentaC(tFilaC *f, const tElemento *item)
{
   ASSEGURA(!FilaCheiaC(f), "Erro: Fila cheia.");

   f->fundo = (f->fundo + 1) % (MAX_ELEMENTOS + 1);

   f->itens[f->fundo] = *item;
}

/****
 *
 * RetiraC(): RetiraC o elemento que se encontra na frente de
 *           uma fila circular
 *
 * Parmetros:
 *      f (entrada/sada) - a fila na qual ser feita a RetiraCda
 *
 * Retorno: O elemento que se encontra na frente da fila
 *
 * Observao: Se a fila estiver vazia, causa o aborto do programa.
 *
 ****/
tElemento RetiraC(tFilaC *f)
{
   ASSEGURA(!FilaVaziaC(f), "Erro: Fila vazia.");

   f->frente = (f->frente + 1) % (MAX_ELEMENTOS + 1);

   return f->itens[f->frente];
}
