/****
 *
 * Arquivo: Fila.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/06/2012
 * ltima modificao: 16/06/2012
 *
 * Descrio: Implementao de fila usando array dinmico linear
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>     /* Funes de Entrada/Sada  */
#include <stdlib.h>    /* Funo exit()             */
#include "Fila.h"      /* Declaraes do tipo fila  */
#include "ASSEGURA.h"  /* Macro ASSEGURA            */

/***************** Constantes Simblicas ******************/

   /* Nmero de elementos que sero acrescentados ao */
   /* array a cada alocao ou realocao de memria */
#define ACRESCIMO 10

/****************** Complemento de Tipo *******************/

struct rotFilaIdxD {
   tItemFila *itens; /* Ponteiro para o array que conter os itens */
   int        frente, fundo;
   int        tamanhoArray; /* Nmero de elementos do array */
};

/***************** Definies de Funes ******************/

/****
 *
 * EstaCheiaFilaIdxD(): Verifica se uma fila est repleta e, se for
 *              o caso, promove o deslocamento de elementos
 *              de modo a criar espao para insero.
 *
 * Parmetros:
 *      f (entrada/sada) - a fila
 *
 * Retorno: 1, se a fila estiver repleta
 *          0, em caso contrrio
 *
 * Observao: Esta funo faz parte da implementao de fila e
 *             no da abstrao fila. Por isso, ela deve ser
 *             definida como static
 *
 ****/
static int EstaCheiaFilaIdxD(tFilaIdxD *f)
{
   int i, deslocamento;

   if ((*f)->fundo < (*f)->tamanhoArray - 1)
      return 0; /* H espao no final do array */

   if ( (*f)->frente == -1 &&
        (*f)->fundo == (*f)->tamanhoArray - 1 )
      return 1; /* No h mais espao no array */

      /* Neste ponto, sabe-se que h espao, mas   */
      /* necessrio mover os elementos da fila para */
      /* o incio do array. O deslocamento a ser    */
      /* feito em cada elemento  dado por:         */
   deslocamento = (*f)->tamanhoArray - ((*f)->fundo -
                  (*f)->frente);

      /* O lao for a seguir desloca cada elemento */
      /* da fila para o incio do array            */
   for (i = (*f)->frente + 1; i <= (*f)->fundo; ++i)
      (*f)->itens[i - deslocamento] = (*f)->itens[i];

      /* A frente e o fundo da fila tambm */
      /* precisam ser deslocados           */
   (*f)->frente -= deslocamento;
   (*f)->fundo -= deslocamento;

      /* A fila foi deslocada e agora     */
      /* podem-se acrescentar novos itens */
   return 0;
}

/****
 *
 * CriaFilaIdxD(): Inicia a fila recebida como parmetro
 *
 * Parmetros:
 *      f (entrada/sada) - a fila
 *
 * Retorno: Nada
 *
 ****/
void CriaFilaIdxD(tFilaIdxD *f)
{
      /* Aloca espao para a estrutura que armazena a fila */
   *f = malloc(sizeof(struct rotFilaIdxD));

      /* Garante que a alocao realmente ocorreu */
   ASSEGURA(*f, "Nao foi possivel alocar fila");

      /* Inicia a frente e o fundo da fila */
   (*f)->frente = (*f)->fundo = -1;

      /* Aloca um array inicial */
   (*f)->itens = calloc(ACRESCIMO, sizeof(tItemFila));

      /* Garante que o array foi realmente alocado */
   ASSEGURA( (*f)->itens, "Nao foi possivel alocar "
             "array para conter elementos da fila" );

      /* Atualiza a varivel que armazena o tamanho do array */
   (*f)->tamanhoArray = ACRESCIMO;
}

/****
 *
 * DestroiFila(): Libera o espao ocupado por uma fila
 *
 * Parmetros: fila (entrada e sada) - a fila
 *
 * Retorno: Nada
 *
 ****/
void DestroiFilaIdxD(tFilaIdxD *fila)
{
      /* Se a fila no estiver vazia, libera o espao */
      /* ocupado pelo array que contm os elementos    */
   if (!EstaVaziaFilaIdxD(fila)) {
      free((*fila)->itens);
   }

      /* Libera a estrutura que representa a fila */
   free((*fila));
}

/****
 *
 * EstaVaziaFilaIdxD(): Verifica se a fila recebida como parmetro
 *              est vazia
 *
 * Parmetros:
 *      f (entrada) - a fila
 *
 * Retorno: 1, se a fila estiver vazia
 *          0, em caso contrrio
 *
 ****/
int EstaVaziaFilaIdxD(const tFilaIdxD *f)
{
   return (*f)->frente == (*f)->fundo;
}

/****
 *
 * ElementoFrenteIdxD(): Retorna o elemento que se encontra
 *                   na frente de uma fila.
 *
 * Parmetros:
 *      f (entrada) - a fila
 *
 * Retorno: O elemento que se encontra na frente da fila.
 *
 * Observao: Se a fila estiver vazia, causa o aborto do programa.
 *
 ****/
tItemFila ElementoFrenteIdxD(const tFilaIdxD *f)
{
   ASSEGURA(!EstaVaziaFilaIdxD(f), "Erro: Fila vazia");

   return (*f)->itens[(*f)->frente + 1];
}

/****
 *
 * EnfileiraIdxD(): Acrescenta um item numa fila.
 *
 * Parmetros:
 *      f (entrada/sada) - a fila na qual ser feita o acrscimo
 *      item (entrada) - o item que ser acrescentado
 *
 * Retorno: Nada.
 *
 * Observao: Se a alocao falhar, causa o aborto do programa
 *
 ****/
void EnfileiraIdxD(tFilaIdxD *f, const tItemFila *item)
{
   tItemFila *novoArray; /* Ponteiro para o array */
                         /* redimensionado        */

      /* Se o array estiver repleto, tenta redimension-lo */
   if (EstaCheiaFilaIdxD(f)) {
         /* Calcula o novo tamanho do array */
      (*f)->tamanhoArray += ACRESCIMO;

         /* Tenta redimensionar o array */
      novoArray = realloc((*f)->itens,
                          (*f)->tamanhoArray*sizeof(tItemFila));

         /* Checa se houve realocao de memria */
      ASSEGURA(novoArray, "Nao houve redimensionamento");

         /* O ponteiro que apontava para o incio */
         /* do array pode no ser mais vlido.    */
         /* Portanto,  preciso atualiz-lo.      */
      (*f)->itens = novoArray;
   }

   ++(*f)->fundo;

   (*f)->itens[(*f)->fundo] = *item;
}

/****
 *
 * DesenfileiraIdxD(): Retira o elemento que se encontra na frente de uma fila
 *
 * Parmetros:
 *      f (entrada/sada) - a fila na qual ser feita a retirada
 *
 * Retorno: O elemento que se encontra na frente da fila
 *
 * Observao: Se a fila estiver vazia, causa o aborto do programa.
 *
 ****/
tItemFila DesenfileiraIdxD(tFilaIdxD  *f)
{
   ASSEGURA(!EstaVaziaFilaIdxD(f), "Erro: Fila vazia.");

   ++(*f)->frente;

   return (*f)->itens[(*f)->frente];
}
