/****
 *
 * Ttulo: Concatenao Mltipla de Strings
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/08/2012
 * ltima modificao: 16/08/2012
 *
 * Descrio: Concatena um string um nmero especificado
 *            de vezes
 *
 * Entrada: O string que ser concatenado
 *
 * Sada: O resultado da concatenao de um string um
 *        nmero especfico de vezes
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>   /* Entrada e sada          */
#include <string.h>  /* Processamento de strings */
#include <stdlib.h>  /* Alocao dinmica        */

/***************** Constantes Simblicas ******************/

   /* Tamanho do acrscimo do bloco usado para conter */
   /* uma linha a cada chamada de realloc()           */
#define TAMANHO_BLOCO 256

/***************** Definies de Funes ******************/

/****
 *
 * LeLinhaIlimitada(): L uma linha de tamanho arbitrrio
 *                     num stream de texto e armazena os
 *                     caracteres lidos num array alocado
 *                     dinamicamente
 *
 * Parmetros:
 *     *tam (sada) - se no for NULL, armazenar o tamanho
 *                    do string constitudo pelos caracteres
 *                    da linha
 *      stream (entrada) - stream de texto no qual ser
 *                         feita a leitura
 *
 * Retorno: Endereo do array contendo a linha lida.
 *          NULL, se ocorrer erro ou o final do arquivo
 *          for atingido antes da leitura de qualquer
 *          caractere
 *
 * Observaes:
 *
 *       1. O stream deve estar associado a um arquivo de
 *          texto aberto em modo de texto que permita
 *          leitura
 *
 *       2. O caractere '\n' no  includo no string
 *          resultante da leitura
 *
 *       3. O primeiro parmetro pode ser NULL. Neste
 *          caso, o tamanho do string no ser armazenado
 *
 ****/
char *LeLinhaIlimitada(int *tam, FILE *stream)
{
   char *ar = NULL,   /* Ponteiro para um array      */
                      /* alocado dinamicamente que   */
                      /* conter os caracteres lidos */
        *p; /* Usado em chamada de realloc() */
   int   tamanho = 0, /* Tamanho do array alocado */
         c,  /* Armazenar cada caractere lido */
         i; /* ndice do prximo caractere */
            /* a ser inserido no array     */

      /* L caracteres a partir da posio corrente do */
      /* apontador de posio do arquivo e armazena-os */
      /* num array. A leitura encerra quando '\n'     */
      /* encontrado, o final do arquivo  atingido ou  */
      /* ocorre erro.                                  */
   for (i = 0; ; ++i) {
         /* L o prximo caractere no arquivo */
      c = fgetc(stream);

         /* Se ocorreu erro de leitura, libera o  */
         /* bloco eventualmente alocado e retorna */
      if (ferror(stream)) {
         free(ar); /* Libera o bloco apontado por 'ar' */
         return NULL; /* Ocorreu erro de leitura */
      }

         /* Verifica se array est completo.      */
         /* Se for o caso, tenta redimension-lo. */
      if (i > tamanho - 2) {
         p = realloc(ar, tamanho + TAMANHO_BLOCO);

            /* Se o redimensionamento no foi     */
            /* possvel, libera o bloco e retorna */
         if (!p) {
            free(ar); /* Libera o bloco apontado por 'ar' */
            return NULL; /* Ocorreu erro de alocao */
         }

            /* Redimensionamento foi OK. Ento, faz-se */
            /* 'ar' apontar para o novo bloco.         */
         ar = p;

            /* O array aumentou de tamanho */
         tamanho = tamanho + TAMANHO_BLOCO;
      }

         /* Se final do arquivo for atingido */
         /* ou o caractere '\n' for lido,    */
         /* encerra a leitura                */
      if (feof(stream) || c == '\n') {
         break; /* Leitura encerrada */
      }

         /* Acrescenta mais um caractere ao array */
      ar[i] = c;
   }

      /* Se nenhum caractere foi lido, libera */
      /* o espao alocado e retorna NULL      */
   if (feof(stream) && !i) {
      free(ar); /* Libera o bloco apontado por 'ar' */
      return NULL; /* Nenhum caractere foi */
                   /* armazenado no array  */
   }

      /* Insere o caractere terminal de string. Neste  */
      /* instante, deve haver espao para ele porque o */
      /* array foi sempre redimensionado deixando um   */
      /* espao para o onipresente caractere '\0'.     */
   ar[i] = '\0';

      /* Atualiza valor apontando pelo parmetro */
      /* 'tam' se ele no for NULL               */
   if (tam) {
      *tam = i; /* i contm o tamanho do array */
   }

   /*                                     */
   /* >>> NB: O tamanho do string no <<< */
   /* >>>     inclui o caractere '\0' <<< */
   /*                                     */

      /* Tenta ajustar o tamanho do array para  */
      /* no haver desperdcio de memria. Note */
      /* que, como i  o tamanho do string, o   */
      /* tamanho do array deve ser i + 1.       */
   p = realloc(ar, i + 1);

   /*                                                 */
   /* Se a realocao foi bem sucedida, retorna-se p. */
   /* Caso contrrio, ar ainda aponta para um bloco   */
   /* vlido. Talvez, haja desperdcio de memria,    */
   /* mas, aqui,  melhor retornar ar do que NULL.    */
   /*                                                 */

   return p ? p : ar;
}

/****
 *
 * LimpaBuffer(): Limpa o buffer associado ao meio de
 *                entrada padro (stdin)
 *
 * Parmetros: Nenhum
 *
 * Retorno: O nmero de caracteres descartados, com
 *          exceo do caractere '\n'.
 *
 ****/
static int LimpaBuffer(void)
{
   int carLido, /* Armazena cada caractere lido */
       nCarLidos = 0; /* Conta o nmero de */
                      /* caracteres lidos  */

      /* L e descarta cada caractere lido at    */
      /* encontrar '\n' ou getchar() retornar EOF */
   do {
      carLido = getchar(); /* L um caractere */
      ++nCarLidos; /* Mais um caractere foi lido */
   } while ((carLido != '\n') && (carLido != EOF));

      /* O ltimo caractere lido foi '\n' ou  */
      /* EOF e no deve ser considerado sobra */
   return nCarLidos - 1;
}

/****
 *
 * LeInteiro(): L um nmero inteiro via teclado
 *              e deixa o buffer intacto
 *
 * Parmetros: Nenhum
 *
 * Retorno: O nmero inteiro lido
 *
 ****/
int LeInteiro(void)
{
   int num, /* O nmero lido */
       teste, /* Valor retornado por scanf() */
       nResto = 0; /* Nmero de caracteres */
                   /* excedentes           */

   /* Desvia para c se o valor for invlido */
inicio:
      /* Tenta ler um valor vlido */
   teste = scanf("%d", &num);

      /* Se no ocorreu erro de leitura ou de final */
      /* de arquivo, h caracteres remanescentes    */
      /* que precisam ser removidos                 */
   if (teste != EOF) {
      nResto = LimpaBuffer();
   }

      /* Enquanto o valor retornado por scanf()  */
      /* no indicar que um valor vlido foi     */
      /* lido continua tentando obter esse valor */
   while(teste != 1) {
      printf("\a\n\t>>> O valor digitado e' invalido. "
                      "Tente novamente\n\t> " );
      goto inicio; /* No causa dano algum */
   }

      /* Repreende o usurio se ele digitou demais */
   if (nResto == 1) {
      printf("\t>>> Um caractere foi descartado\n");
   } else if (nResto > 1) {
      printf("\t>>> %d caracteres foram descartados\n",
             nResto);
   }

      /* O valor retornado certamente  vlido */
   return num;
}

/****
 *
 * ConcatenaNVezes(): Cria um array e armazena nele o
 *                    resultado da concatenao do string
 *                    recebido como parmetro o nmero de
 *                    vezes especificado
 *
 * Parmetros:
 *      n (entrada) - nmero de concatenaes
 *      str (entrada) - string que ser concatenado n vezes
 *
 * Retorno: Endereo de um array alocado dinamicamente
 *          contendo o resultado da concatenao, se
 *          ela for bem sucedida. NULL, em caso contrrio.
 *
 ****/
char *ConcatenaNVezes(int n, const char *str)
{
   char *concatenacao; /* Apontar para o resultado */
   int   i;

      /* O valor de n deve ser pelo menos igual a 1 */
   if (n <= 0) {
      return NULL; /* Por que esta funo foi chamada? */
   }

      /* Tenta alocar o espao necessrio para conter     */
      /* o resultado da concatenao.  necessrio        */
      /* adicionar 1 devido ao onipresente caractere '\0' */
   concatenacao = malloc(n*strlen(str) + 1);

      /* Se o espao no foi alocado, retorna-se NULL */
   if (!concatenacao) {
      return NULL; /* Alocao falhou */
   }

      /* Primeiro, copia o string para o array alocado */
   strcpy(concatenacao, str);

      /* Agora, acrescenta, sucessivamente,  */
      /* n-1 cpias do string ao mesmo array */
   for (i = 1; i < n; i++) {
         /* Acrescenta o string mais uma vez */
      strcat(concatenacao, str);
   }

   return concatenacao;
}

/****
 *
 * main(): Concatena um string especificado pelo usurio
 *         o nmero de vezes escolhido por ele
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero, se no ocorrer nenhum erro.
 *          Um valor diferente de zero em caso contrrio.
 *
 ****/
int main(void)
{
   char *string, /* String que ser concatenado n vezes */
        *resultado; /* Resultado da concatenao */
   int   nVezes; /* Nmero de vezes que o   */
                 /* string ser concatenado */

      /* Apresenta o programa */
   printf( "\n\t>>> Este programa une os caracteres que"
           "\n\t>>> voce especificar o numero de vezes"
           "\n\t>>> que voce desejar, e apresenta o"
           "\n\t>>> resultado na tela.\n" );

      /* L o string que o usurio deseja concatenar */
   printf("\n\t>>> Quais caracteres deseja unir?\n\t> ");
   string = LeLinhaIlimitada(NULL, stdin);

      /* Se o string lido for nulo, o usurio */
      /* no deseja brincar de concatenao   */
   if (!*string) {
      printf("\n\t>>> Nao li nenhum caractere. Tchau.\n");
      return 1; /* O programa no gosta de usurio chato */
   }

      /* Solicita o nmero de concatenaes ao usurio */
   printf( "\n\t>>> Quantas vezes deseja unir os "
           "caracteres? " );

      /* Se o nmero de concatenaes no  */
      /* for aceitvel, encerra o programa */
   if ((nVezes = LeInteiro()) <= 0) {
      printf("\n\t>>> O valor deveria ser pelo menos igual"
             " a 1. Tchau.\n");
      return 1; /* Usurio chato no  tolerado */
   }

      /* Obtm o resultado da concatenao */
   resultado = ConcatenaNVezes( nVezes, string );

      /* Apresenta o resultado da concatenao */
   if (resultado) {
      printf( "\n>>> Resultado da concatenacao de \"%s\" %d"
              " vezes:", string, nVezes );
      printf("\n\n%s\n", resultado);
   } else {
      printf( "\n>>> Nao foi possivel concatenar \"%s\" %d"
              " vezes:", string, nVezes );
   }

      /* Libera os espaos alocados dinamicamente. Isso */
      /* no  necessrio, j que o programa encerrar  */
      /* em seguida, mas para no perder o hbito...    */
   free(string);
   free(resultado);

   return 0;
}

/****************** Resultado do Programa *****************

        >>> Este programa une os caracteres que
        >>> voce especificar o numero de vezes
        >>> que voce desejar, e apresenta o
        >>> resultado na tela.

        >>> Quais caracteres deseja unir?
        > +-

        >>> Quantas vezes deseja unir os caracteres? 20

>>> Resultado da concatenacao de "+-" 20 vezes:

+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-

**********************************************************/
