/****
 *
 * Arquivo: PilhaG.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 01/05/2013
 * ltima modificao: 01/05/2013
 *
 * Descrio: Implementao de operaes com pilhas
 *            genricas
 *
 ****/

/*********************** Includes *************************/

#include <stdlib.h>   /* Alocao dinmica      */
#include <string.h>   /* memcpy()               */
#include "PilhaG.h"   /* Interface deste mdulo */
#include "ASSEGURA.h" /* Macro ASSEGURA         */

/**************** Constantes Simblicas *******************/

   /* Tamanho inicial e de acrscimo do continer */
   /* que armazenar os elementos da pilha        */
#define TAM_INCREMENTO_PILHA 4

/************* Definies de Funes Globais **************/

/****
 *
 * CriaPilhaG(): Cria uma pilha genrica
 *
 * Parmetros:
 *      p (sada) - ponteiro para a pilha que ser iniciada
 *      tamElemento (entrada) - tamanho de cada elemento que
 *                              ser armazenado na pilha
 *
 * Retorno: Nenhum
 *
 * Observao: O programa ser abortado se o segundo
 *             parmetro no for positivo ou se a alocao
 *             de espao reservado aos elementos no for
 *             possvel
 *
 ****/
void CriaPilhaG(tPilhaGen *p, int tamElemento)
{
      /* O tamanho de um elemento deve ser positivo */
   ASSEGURA( tamElemento > 0,
             "Tamanho de elemento deve ser positivo" );

      /* Inicia o campo que indica o tamanho */
      /* de cada elemento da pilha           */
   p->tamElemento = tamElemento;

      /* Inicialmente, a pilha no tem elementos */
   p->nElementos = 0;

      /* Inicia o campo que indica a capacidade da pilha */
   p->capacidade = TAM_INCREMENTO_PILHA;

      /* Aloca o espao inicial do continer da pilha */
   p->elementos = malloc(TAM_INCREMENTO_PILHA*tamElemento);

      /* Se ocorreu falha de alocao, aborta o programa */
   ASSEGURA(p->elementos, "Erro de alocacao");
}

/****
 *
 * DestroiPilhaG(): Libera o espao ocupado pelos elementos
 *                  de uma pilha genrica
 *
 * Parmetros:
 *      p (sada) - ponteiro para a pilha que ter
 *                  espao liberado
 *
 * Retorno: Nenhum
 *
 ****/
void DestroiPilhaG(tPilhaGen *p)
{
      /* Libera o espao ocupado pelo array */
      /* que armazena os elementos da pilha */
   free(p->elementos);

      /* p->elementos deixou de ser um ponteiro vlido */
   p->elementos = NULL;
}

/****
 *
 * PilhaVaziaG(): Verifica se uma pilha genrica est vazia
 *
 * Parmetros:
 *      p (entrada) - ponteiro para a pilha que ser
 *                    verificada
 *
 * Retorno: 1, se a pilha estiver vazia;
 *          0, em caso contrrio
 *
 ****/
int PilhaVaziaG(const tPilhaGen *p)
{
      /* A pilha est vazia quando seu nmero */
      /* de elementos  igual a zero          */
   return p->nElementos == 0;
}

/****
 *
 * EmpilhaG(): Armazena um novo item numa pilha genrica
 *
 * Parmetros:
 *       p (entrada/sada) - ponteiro para a pilha
 *      *pItem (entrada) - contedo do novo item a ser
 *                         empilhado
 *
 * Retorno: Nenhum
 *
 * Observao: Se o novo n da pilha no puder ser alocado,
 *             o programa ser abortado
 *
 ****/
void EmpilhaG(tPilhaGen *p, const void *pItem)
{
   char *pTopo; /* Apontar para o local do empilhamento */

      /* Se a pilha atingiu seu limite de capacidade, */
      /* tenta aumentar o tamanho de seu continer    */
   if (p->nElementos == p->capacidade) {
         /* Calcula a nova capacidade */
      p->capacidade = p->capacidade + TAM_INCREMENTO_PILHA;

         /* Tenta aumentar o tamanho do array que */
         /* armazena os elementos da pilha        */
      p->elementos = realloc(p->elementos,
                             p->capacidade*p->tamElemento);

         /* Se o redimensionamento no foi */
         /* possvel, aborta o programa    */
      ASSEGURA(p->elementos, "Erro de alocacao");
   }

      /* Faz 'pTopo' apontar para o local    */
      /* onde o novo elemento ser empilhado */
   pTopo = (char *)p->elementos +
            p->nElementos*p->tamElemento;

      /* Copia o contedo do novo item */
      /* para o local determinado      */
   memcpy(pTopo, pItem, p->tamElemento);

   p->nElementos++; /* Mais um elemento foi empilhado */
}

/****
 *
 * DesempilhaG(): Remove o elemento do topo de uma
 *                pilha genrica
 *
 * Parmetros:
 *      p (entrada/sada) - ponteiro para a pilha
 *     *pItem (sada) - contedo do elemento que
 *                      se encontra no topo da pilha
 *
 * Retorno: O terceiro parmetro
 *
 * Observao: Se a pilha estiver vazia, o programa ser
 *             abortado
 *
 ****/
void *DesempilhaG(tPilhaGen *p, void *pItem)
{
   const char *pTopo; /* Apontar para o elemento */
                      /* no topo da pilha         */

      /* Se a pilha estiver vazia, aborta o programa */
   ASSEGURA(!PilhaVaziaG(p), "A pilha esta' vazia");

      /* A pilha ficar com um elemento a menos */
   p->nElementos--;

      /* Faz 'pTopo' apontar para o endereo do elemento  */
      /* no topo da pilha. Se o nmero de elementos no   */
      /* tivesse sido decrementado antes, 'pTopo'         */
      /* apontaria para o local onde o prximo elemento   */
      /* seria empilhado (i.e., para um elemento adiante) */
   pTopo = (const char *)p->elementos +
           p->nElementos*p->tamElemento;

      /* Copia o contedo do elemento desempilhado */
   memcpy(pItem, pTopo, p->tamElemento);

   return pItem;
}

/****
 *
 * ElementoTopoG(): Obtm o elemento do topo de uma
 *                  pilha genrica sem desempilh-lo
 *
 * Parmetros:
 *     p (entrada) - ponteiro para a pilha cujo elemento
 *                   do topo ser obtido
 *     *pItem (sada) - contedo do elemento que
 *                      se encontra no topo da pilha
 *
 * Retorno: Ponteiro para o contedo do elemento que se
 *          encontra no topo da pilha
 *
 * Observao: Se a pilha estiver vazia, o programa ser
 *             abortado
 *
 ****/
void *ElementoTopoG(const tPilhaGen *p, void *pItem)
{
   const char *pTopo; /* Apontar para o elemento */
                      /* no topo da pilha         */

      /* Se a pilha estiver vazia, apresenta  */
      /* mensagem de erro e aborta o programa */
   ASSEGURA(!PilhaVaziaG(p), "\nErro: Pilha vazia\n");

      /* Faz 'pTopo' apontar para o */
      /* elemento no topo da pilha  */
   pTopo = (const char *)p->elementos +
          (p->nElementos - 1)*p->tamElemento;

      /* Copia o contedo do elemento que */
      /* se encontra no topo da pilha     */
   memcpy(pItem, pTopo, p->tamElemento);

      /* Retorna o endereo do contedo copiado */
   return pItem;
}
