/****
 *
 * Arquivo: Arvore.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 22/07/2012
 * ltima modificao: 22/07/2012
 *
 * Descrio: Implementao do mdulo Arvore
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>    /* Funes de Entrada/Sada  */
#include <stdlib.h>   /* Funo exit()             */
#include "Arvore.h"   /* Interface deste mdulo    */
#include "ASSEGURA.h" /* Macro ASSEGURA            */

/***************** Definies de Funes ******************/

/****
 *
 * ConstroiNo(): Constri um n de uma rvore binria
 *
 * Parmetros:
 *       item (entrada) - contedo efetivo do n
 *
 * Retorno: Endereo do n criado
 *
 ****/
tArvore ConstroiNo(char item)
{
   tArvore no;

      /* Tenta alocar o novo n */
   no = malloc(sizeof(tNoArvore));

      /* Se no houve alocao, aborta o programa */
   ASSEGURA(no, "Erro: Nao foi possivel alocar no'");

      /* Preenche o contedo do n */
   no->conteudo = item;
   no->esquerda = NULL; /* Este n ainda no tem filho */
   no->direita = NULL;

   return no;
}

/****
 *
 * FilhoEsquerda(): Cria o filho da esquerda de um n de uma
 *                  rvore binria
 *
 * Parmetros:
 *       pNo (entrada/sada) - ponteiro para o n cujo filho
 *                             ser criado
 *       item (entrada) - contedo efetivo do filho
 *
 * Retorno: Nada
 *
 ****/
void FilhoEsquerda(tNoArvore *pNo, char item)
{
      /* O ponteiro para o n no pode ser NULL */
   ASSEGURA(pNo, "Erro: Insercao invalida.");

      /* O n no deve ter filho  esquerda ainda */
   ASSEGURA(!pNo->esquerda, "Erro: Insercao invalida.");

      /* Constri um novo n e faz com que ele seja     */
      /* filho  esquerda do n recebido como parmetro */
   pNo->esquerda = ConstroiNo(item);
}

/****
 *
 * FilhoDireita(): Cria o filho da direita de um n de uma
 *                 rvore binria
 *
 * Parmetros:
 *       pNo (entrada/sada) - ponteiro para o n cujo filho
 *                             ser criado
 *       item (entrada) - contedo efetivo do filho
 *
 * Retorno: Nada
 *
 ****/
void FilhoDireita(tNoArvore *pNo, char item)
{
      /* O ponteiro para o n no pode ser NULL */
   ASSEGURA(pNo, "Erro: Insercao invalida.");

      /* O n no deve ter filho  direita ainda */
   ASSEGURA(!pNo->direita, "Erro: Insercao invalida.");

      /* Constri um novo n e faz com que ele seja    */
      /* filho  direita do n recebido como parmetro */
   pNo->direita = ConstroiNo(item);
}

/****
 *
 * Prefixo(): Efetua um caminhamento em ordem prefixa numa
 *             rvore binria
 *
 * Parmetros:
 *       arvore (entrada) - ponteiro para a raiz da
 *                          rvore sobre a qual ser
 *                          efetuado o caminhamento
 *       op (entrada) - ponteiro para a funo que representa
 *                      a operao a ser efetuada sobre a
 *                      informao contida em cada n
 *
 * Retorno: Nada
 *
 ****/
void Prefixo(tArvore arvore, tOperacao op)
{
   if (arvore) {
         /* Visita a raiz */
      op(arvore->conteudo);

         /* Caminha na subrvore da esquerda */
      Prefixo(arvore->esquerda, op);

         /* Caminha na subrvore da direita */
      Prefixo(arvore->direita, op);
   }
}

/****
 *
 * Infixo(): Efetua um caminhamento em ordem infixa numa
 *            rvore binria
 *
 * Parmetros:
 *       arvore (entrada) - ponteiro para a raiz da
 *                          rvore sobre a qual ser
 *                          efetuado o caminhamento
 *       op (entrada) - ponteiro para a funo que representa
 *                      a operao a ser efetuada sobre a
 *                      informao contida em cada n
 *
 * Retorno: Nada
 *
 ****/
void Infixo(tArvore arvore, tOperacao op)
{
   if (arvore) {
         /* Caminha subrvore esquerda */
      Infixo(arvore->esquerda, op);

      op(arvore->conteudo); /* Visita a raiz */

         /* Caminha subrvore direita */
      Infixo(arvore->direita, op);
   }
}

/****
 *
 * Sufixo(): Efetua um caminhamento em ordem sufixa numa
 *             rvore binria
 *
 * Parmetros:
 *       arvore (entrada) - ponteiro para a raiz da
 *                          rvore sobre a qual ser
 *                          efetuado o caminhamento
 *       op (entrada) - ponteiro para a funo que representa
 *                      a operao a ser efetuada sobre a
 *                      informao contida em cada n
 *
 * Retorno: Nada
 *
 ****/
void Sufixo(tArvore arvore, tOperacao op)
{
   if (arvore) {
         /* Caminha na subrvore esquerda */
      Sufixo(arvore->esquerda, op);

         /* Caminha na subrvore direita */
      Sufixo(arvore->direita, op);

      op(arvore->conteudo); /* Visita raiz */
   }
}

/****
 *
 * ExibeNivel(): Exibe na tela uma rvore por nvel
 *
 * Parmetros:
 *       arvore (entrada) - ponteiro para a raiz da
 *                          rvore que ser impressa
 *       fImpressao (entrada) - ponteiro para a funo que
 *                              efetua a escrita
 *       nivel (entrada) - nvel a partir do qual a rvore
 *                         ser impressa
 *
 * Retorno: Nada
 *
 ****/
static void ExibeNivel( tArvore arvore,
                        tOperacao fImpressao, int nivel )
{
   int i;

   if (arvore) {
      for (i = 0; i < nivel; ++i) {
         printf("\t");
      }

      fImpressao(arvore->conteudo);

      printf("\n");

      ExibeNivel(arvore->direita, fImpressao, nivel + 1);
      ExibeNivel(arvore->esquerda, fImpressao, nivel + 1);
   }
}

/****
 *
 * Exibe(): Exibe na tela uma rvore usando tabulaes
 *         (endentaes) para refletir a hierarquia
 *         entre os ns
 *
 * Parmetros:
 *      arvore (entrada) - ponteiro para a raiz da
 *                         rvore sobre a qual ser
 *                         efetuado o caminhamento
 *      op (entrada) - ponteiro para a funo que representa
 *                     a operao a ser efetuada sobre a
 *                     informao contida em cada n
 *
 * Retorno: Nada
 *
 * Observao: A funo que realmente faz o servio 
 *             ExibeNivel().
 *
 ****/
void Exibe(tArvore arvore, tOperacao fImpressao)
{
   ExibeNivel(arvore, fImpressao, 0);
}

/****
 *
 * DestroiArvoreBin(): Destri uma rvore binria
 *
 * Parmetros:
 *      arvore (entrada/sada) - ponteiro para a raiz
 *                               da rvore binria
 *
 * Retorno: Nada.
 *
 ****/
void DestroiArvoreBin(tArvore arvore)
{
   if (arvore) {
         /* Destri subrvore esquerda */
      DestroiArvoreBin(arvore->esquerda);

         /* Destri subrvore direita */
      DestroiArvoreBin(arvore->direita);

         /* Remove a raiz */
      free(arvore); /* Libera a raiz */
   }
}
