/****
 *
 * Arquivo: Fila.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/06/2012
 * ltima modificao: 16/06/2012
 *
 * Descrio: Implementao de fila usando lista
 *            encadeada
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>    /* Funes de Entrada/Sada     */
#include <stdlib.h>   /* Funes de alocao dinmica */
#include "Fila.h"     /* Interface deste mdulo       */
#include "ASSEGURA.h" /* Macro ASSEGURA               */

/****************** Complemento de Tipo *******************/

typedef  struct noLista {
            tArvore         conteudo;
            struct noLista *proximo;
         } tNoLista, *tLista;

struct rotFila {
   tNoLista *frente, *fundo;
};

/**************** Definies de funes ******************/

/****
 *
 * DestroiListaSE(): Libera o espao ocupado pelos ns de
 *                   uma lista encadeada, tornando-a vazia
 *
 * Parmetros:
 *     lista (entrada/sada) - endereo do ponteiro para a
 *                             lista simplesmente encadeada
 *                             que ser destruda
 *
 * Retorno: Nada
 *
 ****/
static void DestroiListaSE(tLista *lista)
{
   tLista p; /* Aponta para o prximo n a ser liberado */

   if (!*lista) { /* Verifica se a lista est vazia */
      return; /* Lista vazia no precisa ser destruda */
   }

   p = *lista; /* Faz p apontar para o incio da lista */

      /* Visita cada n da lista liberando-o */
   do {
         /* Passa para o prximo n antes que o n corrente */
         /* seja destrudo. No importa que o ponteiro para */
         /*o incio da lista seja alterado. Afinal, este    */
         /* ponteiro ser mesmo anulado.                    */
      *lista = (*lista)->proximo;

      free(p); /* Libera o espao do n corrente */

         /* Faz p apontar para o prximo n, cujo */
         /* endereo est armazenado em *lista    */
      p = *lista;
   } while (p);

      /* Neste ponto, a lista ficou vazia, pois o  */
      /* ltimo valor assumido por *lista foi NULL */
}

/****
 *
 * CriaFila(): Inicia a fila encadeada recebida como parmetro
 *
 * Parmetros:
 *      f (entrada/sada) - a fila
 *
 * Retorno: Nada
 *
 ****/
void CriaFila(tFila *f)
{
   *f = malloc(sizeof(struct rotFila));

   ASSEGURA(*f, "Impossivel alocar fila");

   (*f)->frente = (*f)->fundo = NULL;
}

/****
 *
 * DestroiFila(): Libera o espao ocupado por uma fila
 *
 * Parmetros: fila (entrada e sada) - a fila
 *
 * Retorno: Nada
 *
 ****/
void DestroiFila(tFila *fila)
{
      /* Se a fila no estiver vazia, libera o espao */
      /* ocupado pelo array que contm os elementos    */
   if (!FilaVazia(fila)) {
      DestroiListaSE(&(*fila)->frente);
   }

      /* Libera a estrutura que representa a fila */
   free((*fila));
}

/****
 *
 * FilaVazia(): Verifica se uma fila encadeada est vazia
 *
 * Parmetros:
 *      f (entrada) - a fila
 *
 * Retorno: 1, se a fila estiver vazia
 *          0, em caso contrrio
 *
 ****/
int FilaVazia(const tFila *f)
{
   return !(*f)->frente;
}

/****
 *
 * ElementoFrente(): Retorna o elemento que se encontra
 *                   na frente de uma fila encadeada
 *
 * Parmetros:
 *      f (entrada) - a fila
 *
 * Retorno: O elemento que se encontra na frente da fila
 *
 * Observao: Se a fila estiver vazia, causa o aborto do programa
 *
 ****/
tArvore ElementoFrente(const tFila *f)
{
   ASSEGURA(!FilaVazia(f), "Erro: Fila vazia");

   return (*f)->frente->conteudo;
}

/****
 *
 * Acrescenta(): Acrescenta um item numa fila encadeada.
 *
 * Parmetros:
 *      f (entrada/sada) - a fila na qual ser feita o acrscimo
 *      item (entrada) - o item que ser acrescentado
 *
 * Retorno: Nada.
 *
 * Observao: Se no for possvel alocar dinamicamente um novo
 *             n, causa o aborto do programa.
 *
 ****/
void Acrescenta(tFila *f, const tArvore *item)
{
   tNoLista  *aux;

      /* Tenta alocar um novo n e causa */
      /* aborto se isto no for possvel */
   ASSEGURA( aux = malloc(sizeof(tNoLista)),
             "Erro de alocao de memria." );

   aux->conteudo = *item; /* preenche o contedo do n */

      /* Este ser o ltimo elemento da fila */
   aux->proximo = NULL;

         /* Se a fila no estiver vazia, o ltimo elemento */
         /* atual deve apontar para o novo elemento. Se a  */
         /* fila estiver vazia, a frente tambm deve       */
         /* apontar para este elemento                     */
   if (FilaVazia(f))
      (*f)->frente = aux;
   else
      (*f)->fundo->proximo = aux;

      /* O fundo deve apontar para o novo elemento */
   (*f)->fundo = aux;
}

/****
 *
 * Retira(): Retira o elemento que se encontra na frente de
 *           uma fila encadeada
 *
 * Parmetros:
 *      f (entrada/sada) - a fila na qual ser feita a retirada
 *
 * Retorno: O elemento que se encontra na frente da fila
 *
 * Observao: Se a fila estiver vazia, causa o aborto do programa.
 *
 ****/
tArvore Retira(tFila *f)
{

   tArvore  elementoNaFrente;
   tNoLista        *aux;

   ASSEGURA(!FilaVazia(f), "Erro: Fila vazia.");

      /* Guarda valor a ser retornado */
   elementoNaFrente = (*f)->frente->conteudo;

      /* Guarda endereo do elemento a ser removido */
   aux = (*f)->frente;

      /* Faz a frente apontar para o elemento seguinte */
   (*f)->frente = (*f)->frente->proximo;

   free(aux); /* Libera o elemento removido */

   return elementoNaFrente;
}
