/****
 *
 * Arquivo: Arvore.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 22/07/2012
 * ltima modificao: 22/07/2012
 *
 * Descrio: Implementao do mdulo Arvore
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>    /* Funes de Entrada/Sada  */
#include <stdlib.h>   /* Funo exit()             */
#include "Arvore.h"   /* Interface deste mdulo    */
#include "ASSEGURA.h" /* Macro ASSEGURA            */

/***************** Definies de Funes ******************/

/****
 *
 * ConstroiNoArvoreBin(): Constri um n de uma rvore binria
 *
 * Parmetros:
 *       item (entrada) - contedo efetivo do n
 *
 * Retorno: Endereo do n criado
 *
 ****/
tArvoreBin ConstroiNoArvoreBin(tConteudo item)
{
   tArvoreBin no;

   no = malloc(sizeof(tNoArvoreBin)); /* Tenta alocar o novo n */

      /* Se no houve alocao, aborta o programa */
   ASSEGURA(no, "Erro: Nao foi possivel alocar no'");

      /* Preenche o contedo do n */
   no->conteudo = item;
   no->esquerda = NULL; /* Este n ainda no tem filho */
   no->direita = NULL;

   return no;
}

/****
 *
 * FilhoEsquerdo(): Cria o filho da esquerda de um n de uma rvore binria
 *
 * Parmetros:
 *       pNo (entrada/sada) - ponteiro para o n cujo filho ser criado
 *       item (entrada) - contedo efetivo do filho
 *
 * Retorno: Nada
 *
 ****/
void FilhoEsquerdo(tNoArvoreBin *pNo, tConteudo item)
{
      /* O ponteiro para o n no pode ser NULL */
   ASSEGURA(pNo, "Erro: Insercao invalida.");

      /* O n no deve ter filho  esquerda ainda */
   ASSEGURA(!pNo->esquerda, "Erro: Insercao invalida.");

      /* Constri um novo n e faz com que ele seja     */
      /* filho  esquerda do n recebido como parmetro */
   pNo->esquerda = ConstroiNoArvoreBin(item);
}

/****
 *
 * FilhoDireito(): Cria o filho da direita de um n de uma rvore binria
 *
 * Parmetros:
 *       pNo (entrada/sada) - ponteiro para o n cujo filho ser criado
 *       item (entrada) - contedo efetivo do filho
 *
 * Retorno: Nada
 *
 ****/
void FilhoDireito(tNoArvoreBin *pNo, tConteudo item)
{
      /* O ponteiro para o n no pode ser NULL */
   ASSEGURA(pNo, "Erro: Insercao invalida.");

      /* O n no deve ter filho  direita ainda */
   ASSEGURA(!pNo->direita, "Erro: Insercao invalida.");

      /* Constri um novo n e faz com que ele seja    */
      /* filho  direita do n recebido como parmetro */
   pNo->direita = ConstroiNoArvoreBin(item);
}

/****
 *
 * CaminhamentoPrefixo(): Efetua um caminhamento em ordem prefixa numa
 *                        rvore binria
 *
 * Parmetros:
 *       arvore (entrada) - ponteiro para a raiz da rvore sobre a qual ser
 *                          efetuado o caminhamento
 *       op (entrada) - ponteiro para a funo que representa a operao a ser
 *                      efetuada sobre a informao contida em cada n
 *
 * Retorno: Nada
 *
 ****/
void CaminhamentoPrefixo(tArvoreBin arvore, tOperacao op)
{
   if (arvore) {
      op(arvore->conteudo); /* Visita a raiz */

         /* Caminha na subrvore da esquerda */
      CaminhamentoPrefixo(arvore->esquerda, op);

         /* Caminha na subrvore da direita */
      CaminhamentoPrefixo(arvore->direita, op);
   }
}

/****
 *
 * CaminhamentoInfixo(): Efetua um caminhamento em ordem infixa numa
 *                       rvore binria
 *
 * Parmetros:
 *       arvore (entrada) - ponteiro para a raiz da rvore sobre a qual ser
 *                          efetuado o caminhamento
 *       op (entrada) - ponteiro para a funo que representa
 *                      a operao a ser efetuada sobre a
 *                      informao contida em cada n
 *
 * Retorno: Nada
 *
 ****/
void CaminhamentoInfixo(tArvoreBin arvore, tOperacao op)
{
   if (arvore) {
      CaminhamentoInfixo(arvore->esquerda, op); /* Caminha subrvore esquerda */

      op(arvore->conteudo); /* Visita a raiz */

      CaminhamentoInfixo(arvore->direita, op); /* Caminha subrvore direita */
   }
}

/****
 *
 * CaminhamentoSufixo(): Efetua um caminhamento em ordem sufixa numa rvore binria
 *
 * Parmetros:
 *       arvore (entrada) - ponteiro para a raiz da rvore sobre a qual ser
 *                          efetuado o caminhamento
 *       op (entrada) - ponteiro para a funo que representa
 *                      a operao a ser efetuada sobre a
 *                      informao contida em cada n
 *
 * Retorno: Nada
 *
 ****/
void CaminhamentoSufixo(tArvoreBin arvore, tOperacao op)
{
   if (arvore) {
      CaminhamentoSufixo(arvore->esquerda, op); /* Caminha na subrvore esquerda */

      CaminhamentoSufixo(arvore->direita, op); /* Caminha na subrvore direita */

      op(arvore->conteudo); /* Visita raiz */
   }
}

/****
 *
 * ExibeNivel(): Exibe na tela uma rvore por nvel
 *
 * Parmetros:
 *       arvore (entrada) - ponteiro para a raiz da rvore que ser impressa
 *       fImpressao (entrada) - ponteiro para a funo que efetua a escrita
 *       nivel (entrada) - nvel a partir do qual a rvore ser impressa
 *
 * Retorno: Nada
 *
 ****/
static void ExibeNivel( tArvoreBin arvore, tOperacao fImpressao, int nivel )
{
   int i;

   if (arvore) {
      for (i = 0; i < nivel; ++i)
         printf("\t");

      fImpressao(arvore->conteudo);

      printf("\n");

      ExibeNivel(arvore->direita, fImpressao, nivel + 1);
      ExibeNivel(arvore->esquerda, fImpressao, nivel + 1);
   }
}

/****
 *
 * ExibeArvoreBin(): Exibe na tela uma rvore usando tabulaes
 *                   (endentaes) para refletir a hierarquia entre os ns
 *
 * Parmetros:
 *      arvore (entrada) - ponteiro para a raiz da rvore sobre a qual ser
 *                         efetuado o caminhamento
 *      op (entrada) - ponteiro para a funo que representa
 *                     a operao a ser efetuada sobre a
 *                     informao contida em cada n
 *
 * Retorno: Nada
 *
 * Observao: A funo que realmente faz o servio  ExibeNivel()
 *
 ****/
void ExibeArvoreBin(tArvoreBin arvore, tOperacao fImpressao)
{
   ExibeNivel(arvore, fImpressao, 0);
}

/****
 *
 * DestroiArvoreBin(): Destri uma rvore binria
 *
 * Parmetros:
 *      arvore (entrada/sada) - ponteiro para a raiz da rvore binria
 *
 * Retorno: Nada.
 *
 ****/
void DestroiArvoreBin(tArvoreBin arvore)
{
   if (arvore) {
      DestroiArvoreBin(arvore->esquerda); /* Destri subrvore esquerda */

      DestroiArvoreBin(arvore->direita); /* Destri subrvore direita */

      free(arvore); /* Remove a raiz */
   }
}
