/****
 *
 * Arquivo: Arvore2.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 22/07/2012
 * ltima modificao: 22/07/2012
 *
 * Descrio: Implementao do mdulo Arvore2
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>    /* Funes de entrada e sada  */
#include "Arvore.h"   /* Definies do tipo rvore   */
#include "ASSEGURA.h" /* Macro ASSEGURA              */

/***************** Definies de Funes ******************/

/****
 *
 * ConstroiNoArvoreBin(): Constri um n de uma rvore binria
 *
 * Parmetros:
 *      item (entrada) - item de dados do n
 *
 * Retorno: Endereo do n criado
 *
 ****/
tArvoreBin ConstroiNoArvoreBin(tConteudo item)
{
   tArvoreBin no;

   no = malloc(sizeof(tNoArvoreBin));

   ASSEGURA(no, "Erro: Nao foi possivel alocar no'");

   no->conteudo = item;
   no->esquerda = NULL;
   no->direita = NULL;

   return no;
}

/****
 *
 * FilhoEsquerdo(): Cria o filho da esquerda de um n de uma
 *                  rvore binria
 *
 * Parmetros:
 *       pNo (entrada e sada) - ponteiro para o n cujo filho
 *                             ser criado
 *      item (entrada) - item de dados do filho
 *
 * Retorno: Nada
 *
 ****/
void FilhoEsquerdo(tNoArvoreBin *pNo, tConteudo item)
{
     /* O ponteiro para o n no pode ser NULL */
   ASSEGURA(pNo, "Erro: Insercao invalida.");

     /* O n no deve ter filho  esquerda ainda */
   ASSEGURA(!pNo->esquerda, "Erro: Insercao invalida.");

   pNo->esquerda = ConstroiNoArvoreBin(item);
}

/****
 *
 * FilhoDireito(): Cria o filho da direita de um n de uma
 *                 rvore binria
 *
 * Parmetros:
 *       pNo (entrada e sada) - ponteiro para o n cujo filho
 *                             ser criado
 *      item (entrada) - item de dados do filho
 *
 * Retorno: Nada
 *
 ****/
void FilhoDireito(tNoArvoreBin *pNo, tConteudo item)
{
     /* O ponteiro para o n no pode ser NULL */
   ASSEGURA(pNo, "Erro: Insercao invalida.");

     /* O n no deve ter filho  direita ainda */
   ASSEGURA(!pNo->direita, "Erro: Insercao invalida.");

   pNo->direita = ConstroiNoArvoreBin(item);
}

/****
 *
 * CaminhamentoPrefixo(): Efetua um caminhamento em ordem prefixa numa
 *             rvore binria
 *
 * Parmetros:
 *      arvore (entrada) - ponteiro para a raiz da
 *                         rvore sobre a qual ser
 *                         efetuado o caminhamento
 *      op (entrada) - ponteiro para a funo que representa
 *                     a operao a ser efetuada sobre a
 *                     informao contida em cada n
 *
 * Retorno: Nada
 *
 ****/
void CaminhamentoPrefixo(tArvoreBin arvore, tOperacao op)
{
   if (arvore) {
       /* Visita a raiz */
     op(arvore->conteudo);

       /* Caminha na subrvore da esquerda */
     CaminhamentoPrefixo(arvore->esquerda, op);

       /* Caminha na subrvore da direita */
     CaminhamentoPrefixo(arvore->direita, op);
   }
}

/****
 *
 * ExibeNivel(): Exibe na tela uma rvore por nvel
 *
 * Parmetros:
 *       arvore (entrada) - ponteiro para a raiz da
 *                          rvore que ser impressa
 *       fImpressao (entrada) - ponteiro para a funo que
 *                              efetua a escrita
 *       nivel (entrada) - nvel a partir do qual a rvore
 *                         ser impressa
 *
 * Retorno: Nada
 *
 ****/
static void ExibeNivel( tArvoreBin arvore,
                        tOperacao fImpressao, int nivel )
{
   int i;

   if (arvore) {
      for (i = 0; i < nivel; ++i) {
         printf("\t");
      }

      fImpressao(arvore->conteudo);

      printf("\n");

      ExibeNivel(arvore->direita, fImpressao, nivel + 1);
      ExibeNivel(arvore->esquerda, fImpressao, nivel + 1);
   }
}

/****
 *
 * ExibeArvoreBin(): Exibe na tela uma rvore usando tabulaes
 *         (endentaes) para refletir a hierarquia
 *         entre os ns
 *
 * Parmetros:
 *      arvore (entrada) - ponteiro para a raiz da
 *                         rvore sobre a qual ser
 *                         efetuado o caminhamento
 *      op (entrada) - ponteiro para a funo que representa
 *                     a operao a ser efetuada sobre a
 *                     informao contida em cada n
 *
 * Retorno: Nada
 *
 * Observao: A funo que realmente faz o servio 
 *             ExibeNivel().
 *
 ****/
void ExibeArvoreBin(tArvoreBin arvore, tOperacao fImpressao)
{
   ExibeNivel(arvore, fImpressao, 0);
}

/****
 *
 * AlturaArvoreBin(): Calcula a altura de uma rvore
 *                 binria
 *
 * Parmetros:
 *      p (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: A referida altura
 *
 ****/
int AlturaArvoreBin(tArvoreBin p)
{
   int profEsq, profDir;

   if (!p) {
      return 0;
   } else {
        /* Calcula a altura da subrvore esquerda */
      profEsq = 1 + AlturaArvoreBin(p->esquerda);

        /* Calcula a altura da subrvore direita */
      profDir = 1 + AlturaArvoreBin(p->direita);

        /* A altura da rvore  a maior */
      return profEsq > profDir ? profEsq : profDir;
   }
}

/****
 *
 * SaoSemelhantesArvoresBin(): Verifica se duas rvores binrias so
 *                   semelhantes
 *
 * Parmetros:
 *      arvore1, arvore2 (entrada) - ponteiros para as
 *                                   razes das rvores que
 *                                   sero comparadas
 *
 * Retorno: 1, se as rvores so semelhantes;
 *          0, em caso contrrio
 *
 * Nota: Duas rvores binrias so semelhantes se elas
 *       so ambas vazias ou suas subrvores esquerdas
 *       so semelhantes e suas subrvores direitas so
 *       semelhantes
 *
 ****/
int SaoSemelhantesArvoresBin(tArvoreBin arvore1, tArvoreBin arvore2)
{
   if (!arvore1 && !arvore2) {
      return 1;
   } else if (arvore1 && arvore2) {
      return SaoSemelhantesArvoresBin( arvore1->esquerda,
                             arvore2->esquerda ) &&
             SaoSemelhantesArvoresBin( arvore1->direita,
                             arvore2->direita );
   } else {
      return 0;
   }
}

/****
 *
 * CopiaArvoreBin(): Faz uma cpia da rvore binria recebida
 *                como parmetro
 *
 * Parmetros:
 *      arvore (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: Endereo da raiz da rvore criada
 *
 ****/
tArvoreBin CopiaArvoreBin(tArvoreBin arvore)
{
   tArvoreBin copia;

   if (!arvore) {
         /* A rvore  vazia e no h o que copiar */
      return NULL;
   } else {
         /* Cria a raiz da rvore */
      copia = ConstroiNoArvoreBin(arvore->conteudo);

         /* Chama esta funo recursivamente para copiar */
         /* os filhos da esquerda e da direita           */
      copia->esquerda = CopiaArvoreBin(arvore->esquerda);
      copia->direita = CopiaArvoreBin(arvore->direita);

      return copia;
   }
}

/****
 *
 * NivelDeNoArvoreBin(): Retorna o nvel de um n de uma rvore
 *              binria
 *
 * Parmetros:
 *      arvore (entrada) - ponteiro para a raiz da rvore
 *      pNo (entrada) - ponteiro para o n
 *
 * Retorno: O nvel do referido n
 *
 ****/
int NivelDeNoArvoreBin(tArvoreBin arvore, tArvoreBin pNo)
{
   int nEsq, nDir;

   if (!arvore) {
      return 0;
   } else if (arvore == pNo) {
      return 1;
   } else {
      nEsq = NivelDeNoArvoreBin(arvore->esquerda, pNo);
      nDir = NivelDeNoArvoreBin(arvore->direita, pNo);

      if (nEsq) {
         return 1 + nEsq;
      } else if (nDir) {
         return 1 + nDir;
      } else {
         return 0;
      }
   }
}

/****
 *
 * CriaArvoreFibonacci(): Retorna um ponteiro para a rvore binria
 *              de Fibonacci de ordem n
 *
 * Parmetros:
 *      arvore (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: O nmero de ns da rvore
 *
 * Nota: Numa rvore binria de Fibonacci de ordem n, se
 *       n = 0 ou n = 1, a rvore consiste de um nico n;
 *       se n > 1, a rvore consiste de uma raiz, tendo a
 *       rvore de Fibonacci de ordem n - 1 como sua
 *       subrvore esquerda e a rvore de Fibonacci de
 *       ordem n - 2 como sua subrvore direita.
 *
 ****/
tArvoreBin CriaArvoreFibonacci(int n)
{
   tArvoreBin fib;

   if (n <= 0) {
      return NULL;
   }

   if (n <= 1) {
      return ConstroiNoArvoreBin(n);
   } else {
      fib = ConstroiNoArvoreBin(n);
      fib->esquerda = CriaArvoreFibonacci(n - 1);
      fib->direita = CriaArvoreFibonacci(n - 2);

      return fib;
   }

   return NULL; /* S para agradar o compilador! */
}

/****
 *
 * NumeroDeNos(): Calcula o nmero de ns de uma
 *                rvore binria
 *
 * Parmetros:
 *      arvore (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: O nmero de ns da rvore
 *
 ****/
int NumeroDeNos(tArvoreBin arvore)
{
   if (!arvore) {
         /* O nmero de ns de uma rvore vazia  0 */
      return 0;
   } else {
         /* O nmero de ns de uma rvore no vazia  */
         /* igual a 1 mais o nmero de ns de sua     */
         /* subrvore esquerda mais o nmero de ns   */
         /* de sua subrvore direita                  */
      return 1 + NumeroDeNos(arvore->esquerda)
               + NumeroDeNos(arvore->direita);
   }
}

/****
 *
 * DestroiArvoreBin(): Destri uma rvore binria
 *
 * Parmetros:
 *      arvore (entrada e sada) - ponteiro para a raiz da
 *                                 rvore binria
 *
 * Retorno: Nada
 *
 ****/
void DestroiArvoreBin(tArvoreBin arvore)
{
   if (arvore) {
         /* Destri subrvore esquerda */
      DestroiArvoreBin(arvore->esquerda);

         /* Destri subrvore direita */
      DestroiArvoreBin(arvore->direita);

         /* Remove a raiz */
      free(arvore); /* Libera a raiz */
   }
}
